@@ -1,7 +1,7 @@
 /*************************** wnchyppr.cpp **********************************
 * Author:        Agner Fog
 * Date created:  2002-10-20
-* Last modified: 2013-11-06
+* Last modified: 2023-05-31
 * Project:       stocc.zip
 * Source URL:    www.agner.org/random
 *
@@ -20,8 +20,8 @@
 * The file ran-instructions.pdf contains further documentation and 
 * instructions.
 *
-* Copyright 2002-2013 by Agner Fog. 
-* GNU General Public License http://www.gnu.org/licenses/gpl.html
+* Copyright 2002-2023 by Agner Fog. 
+* GNU General Public License v3. 3. http://www.gnu.org/licenses/gpl.html
 *****************************************************************************/
 
 #include <string.h>                    // memcpy function
@@ -49,7 +49,7 @@ Log and Exp functions with special care for small x
 // Assembly library randomaXX.lib is used.
 // Nothing to include here.
 
-#elif defined(__GNUC__) || defined (__INTEL_COMPILER) || defined(HAVE_EXPM1)
+#elif defined(__GNUC__) || defined (__clang__) || defined(__INTEL_COMPILER) || defined(HAVE_EXPM1)
 // (2) 
 // Functions log1p(x) = log(1+x) and expm1(x) = exp(x)-1 are available
 // in the math libraries of Gnu and Intel compilers 
@@ -109,7 +109,7 @@ double log1pow(double q, double x) {
 
 #else
 // (3)
-// Functions log1p and expm1 are not available in MS and Borland compiler
+// Functions log1p and expm1 are not available in old MS and Borland compiler
 // libraries. Use explicit Taylor expansion when needed.
 
 double pow2_1(double q, double * y0 = 0) {
@@ -340,7 +340,8 @@ Methods for class CWalleniusNCHypergeometric
 CWalleniusNCHypergeometric::CWalleniusNCHypergeometric(int32 n_, int32 m_, int32 N_, double odds_, double accuracy_) {
    // constructor
    accuracy = accuracy_;
-   SetParameters(n_, m_, N_, odds_);}
+   SetParameters(n_, m_, N_, odds_);
+}
 
 
 void CWalleniusNCHypergeometric::SetParameters(int32 n_, int32 m_, int32 N_, double odds) {
@@ -633,8 +634,8 @@ int CWalleniusNCHypergeometric::BernouilliH(int32 x_, double h, double rh, Stoch
    // This function is intended for use in rejection methods for sampling from
    // the Wallenius distribution. It is called from 
    // StochasticLib3::WalleniusNCHypRatioOfUnifoms in the file stoc3.cpp
-   double f0;                      // Lambda*Phi()
-   double phideri0;                // phi()/rd
+   double f0;                      // Lambda*Phi(0.5)
+   double phideri0;                // phi(0.5)/rd
    double qi;                      // 2^(-r*omega[i])
    double qi1;                     // 1-qi
    double omegai[2] = {omega,1.};  // weights for each color
@@ -665,7 +666,7 @@ int CWalleniusNCHypergeometric::BernouilliH(int32 x_, double h, double rh, Stoch
    k *= w;                         // w * k   
    rdm1 = rd - 1.;
 
-   // calculate phi()/rd
+   // calculate phi(0.5)/rd
    phideri0 = -LN2 * rdm1;
    for (i=0; i<2; i++) {
       romegi = r * omegai[i];
@@ -716,50 +717,60 @@ double CWalleniusNCHypergeometric::recursive() {
    // Wallenius noncentral hypergeometric distribution by recursion formula
    // Approximate by ignoring probabilities < accuracy and minimize storage requirement
    const int BUFSIZE = 512;            // buffer size
-   double p[BUFSIZE+2];                // probabilities
-   double * p1, * p2;                  // offset into p
+   double pp[BUFSIZE+2];               // probabilities
+   //double * p1, * p2;                // offset into pp
+   int32 j1, j2;                       // offset into pp
+   /* pointer arithmetics in p1, p2 in earlier versions replaced by offset j1, j2
+      because of false error messages by gcc-UBSAN   */
+
    double mxo;                         // (m-x)*omega
    double Nmnx;                        // N-m-nu+x
    double y, y1;                       // save old p[x] before it is overwritten
-   double d1, d2, dcom;                // divisors in probability formula
+   double d1, d2;                      // divisors in probability formula
    double accuracya;                   // absolute accuracy
    int32 xi, nu;                       // xi, nu = recursion values of x, n
    int32 x1, x2;                       // xi_min, xi_max
 
-   accuracya = 0.005f * accuracy;      // absolute accuracy
-   p1 = p2 = p + 1;                    // make space for p1[-1]
-   p1[-1] = 0.;  p1[0]  = 1.;          // initialize for recursion
+   accuracya = 0.005 * accuracy;       // absolute accuracy
+   j1 = j2 = 1;                        // make space for pp[j1-1]
+   pp[0] = 0.;  pp[1] = 1.;            // initialize for recursion
    x1 = x2 = 0;
-   for (nu=1; nu<=n; nu++) {
-      if (n - nu < x - x1 || p1[x1] < accuracya) {
+   for (nu = 1; nu <= n; nu++) {
+      //if (j1+x1 < 0 || j1+x2 < 0) FatalError("j1+x1 < 0");
+      if (n - nu < x - x1 || pp[j1+x1] < accuracya) {
          x1++;               // increase lower limit when breakpoint passed or probability negligible
-         p2--;               // compensate buffer offset in order to reduce storage space
+         j2--;               // compensate buffer offset in order to reduce storage space
       }
-      if (x2 < x && p1[x2] >= accuracya) {
+      if (x2 < x && pp[j1+x2] >= accuracya) {
          x2++;  y1 = 0.;     // increase upper limit until x has been reached
       }
       else {
-         y1 = p1[x2];
+         y1 = pp[j1+x2];
       }
       if (x1 > x2) return 0.;
-      if (p2+x2-p > BUFSIZE) FatalError("buffer overrun in function CWalleniusNCHypergeometric::recursive");
+      if (j2+x2 > BUFSIZE) FatalError("buffer overrun in function CWalleniusNCHypergeometric::recursive");
 
       mxo = (m-x2)*omega;
       Nmnx = N-m-nu+x2+1;
       for (xi = x2; xi >= x1; xi--) {    // backwards loop
+         //if (j1+xi < 1 || j2+xi < 1) FatalError("j1+xi < 1");
          d2 = mxo + Nmnx;
          mxo += omega; Nmnx--;
          d1 = mxo + Nmnx;
-         dcom = 1. / (d1 * d2);           // save a division by making common divisor
-         y  = p1[xi-1]*mxo*d2*dcom + y1*(Nmnx+1)*d1*dcom;
-         y1 = p1[xi-1];                   // (warning: pointer alias, can't swap instruction order)
-         p2[xi] = y;
+         // save a division by making common divisor
+         //dcom = 1. / (d1 * d2);
+         //y  = pp[j1+xi-1]*mxo*d2*dcom + y1*(Nmnx+1)*d1*dcom;
+         y  = (pp[j1+xi-1]*mxo*d2 + y1*(Nmnx+1)*d1) / (d1 * d2);
+         y1 = pp[j1+xi-1];
+         pp[j2+xi] = y;
       }
-      p1 = p2;
+      j1 = j2;
    }
 
    if (x < x1 || x > x2) return 0.;
-   return p1[x];
+   //if (j1+x < 0) FatalError("j1+x < 0");
+
+   return pp[j1+x];
 }
 
 
@@ -1201,7 +1212,7 @@ double CWalleniusNCHypergeometric::probability(int32 x_) {
 }
 
 
-int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int32 * xfirst, int32 * xlast, double cutoff) {
+int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int32 * xfirst, int32 * xlast, bool * useTable, double cutoff) {
    // Makes a table of Wallenius noncentral hypergeometric probabilities 
    // table must point to an array of length MaxLength. 
    // The function returns 1 if table is long enough. Otherwise it fills
@@ -1216,9 +1227,9 @@ int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int
    //
    // The function will return the following information when MaxLength = 0:
    // The return value is the desired length of table.
-   // *xfirst is 1 if it will be more efficient to call MakeTable than to call
+   // useTable is true if it will be more efficient to call MakeTable than to call
    // probability repeatedly, even if only some of the table values are needed.
-   // *xfirst is 0 if it is more efficient to call probability repeatedly.
+   // useTable is false if it is more efficient to call probability repeatedly.
 
    double * p1, * p2;                  // offset into p
    double mxo;                         // (m-x)*omega
@@ -1229,8 +1240,8 @@ int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int
    int32 xi, nu;                       // xi, nu = recursion values of x, n
    int32 x1, x2;                       // lowest and highest x or xi
    int32 i1, i2;                       // index into table
-   int32 UseTable;                     // 1 if table method used
-   int32 LengthNeeded;                 // Necessary table length
+   bool  useTabl;                      // true if table method used
+   int32 lengthNeeded;                 // Necessary table length
 
    // special cases
    if (n == 0 || m == 0) {x1 = 0; goto DETERMINISTIC;}
@@ -1240,28 +1251,30 @@ int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int
       if (n > N-m) FatalError("Not enough items with nonzero weight in  CWalleniusNCHypergeometric::MakeTable");
       x1 = 0;
       DETERMINISTIC:
-      if (MaxLength == 0) {
-         if (xfirst) *xfirst = 1;
-         return 1;
-      }
       *xfirst = *xlast = x1;
-      *table = 1.;
+      if (MaxLength && table) *table = 1.;
+      if (useTable) *useTable = true;
       return 1;
    }
+   
+   int32 L = n + m - N;                // parameter
+   x1 = (L > 0) ? L : 0;               // xmin
+   x2 = (n < m) ? n : m;               // xmax
+   *xfirst = x1;  *xlast = x2;   
 
    if (cutoff <= 0. || cutoff > 0.1) cutoff = 0.01 * accuracy;
 
-   LengthNeeded = N - m;               // m2
-   if (m < LengthNeeded) LengthNeeded = m;
-   if (n < LengthNeeded) LengthNeeded = n; // LengthNeeded = min(m1,m2,n)
-   area = double(n)*LengthNeeded;      // Estimate calculation time for table method
-   UseTable = area < 5000. || (area < 10000. && N > 1000. * n);
+   lengthNeeded = N - m;               // m2
+   if (m < lengthNeeded) lengthNeeded = m;
+   if (n < lengthNeeded) lengthNeeded = n; // lengthNeeded = min(m1,m2,n)
+   area = double(n)*lengthNeeded;      // Estimate calculation time for table method
+   useTabl = area < 5000. || (area < 10000. && N > 1000. * n);
+   if (useTable) *useTable = useTabl;
 
    if (MaxLength <= 0) {
-      // Return UseTable and LengthNeeded
-      if (xfirst) *xfirst = UseTable;
-      i1 = LengthNeeded + 2;           // Necessary table length
-      if (!UseTable && i1 > 200) {
+      // Return useTabl and lengthNeeded
+      i1 = lengthNeeded + 2;           // Necessary table length
+      if (!useTabl && i1 > 200) {
          // Calculate necessary table length from standard deviation
          double sd = sqrt(variance()); // calculate approximate standard deviation
          // estimate number of standard deviations to include from normal distribution
@@ -1271,7 +1284,7 @@ int32 CWalleniusNCHypergeometric::MakeTable(double * table, int32 MaxLength, int
       return i1;
    }
 
-   if (UseTable && MaxLength > LengthNeeded) {
+   if (useTabl && MaxLength > lengthNeeded) {
       // use recursion table method
       p1 = p2 = table + 1;             // make space for p1[-1]
       p1[-1] = 0.;  p1[0] = 1.;        // initialize for recursion
